# these are all just exports, #noqa them so flake8 will be happy
from dbt.fal.adapters.python.connections import PythonConnectionManager  # noqa
from dbt.fal.adapters.python.impl import PythonAdapter  # noqa



from typing import Any, Optional, Type, Set

from dbt.adapters.factory import get_adapter_by_type
from dbt.adapters.base.meta import available
from dbt.adapters.base.impl import BaseAdapter
from dbt.contracts.connection import Credentials
from dbt.parser.manifest import ManifestLoader
from dbt.clients.jinja import MacroGenerator

from ..fal_experimental.utils import cache_static
from ..fal_experimental.impl import FalAdapterMixin
from ..fal_experimental import telemetry


class FalCredentialsWrapper:
    _db_creds: Optional[Credentials] = None

    def __init__(self, db_creds: Credentials):
        self._db_creds = db_creds

    @property
    def type(self):
        if find_funcs_in_stack({"to_target_dict", "db_materialization"}):
            # This makes sense for both SQL and Python because the target is always the db
            return self._db_creds.type

        return "fal"

    def __getattr__(self, name: str) -> Any:
        """
        Directly proxy to the DB adapter, just shadowing the type
        """
        return getattr(self._db_creds, name)


class FalEncAdapterWrapper(FalAdapterMixin):
    def __init__(self, db_adapter_type: Type[BaseAdapter], config):
        # Use the db_adapter_type connection manager
        self.ConnectionManager = db_adapter_type.ConnectionManager

        db_adapter = get_adapter_by_type(db_adapter_type.type())
        super().__init__(config, db_adapter)

        # HACK: A Python adapter does not have self._available_ all the attributes a DB adapter does.
        # Since we use the DB adapter as the storage for the Python adapter, we must proxy to it
        # all the unhandled calls.

        # self._available_ is set by metaclass=AdapterMeta
        self._available_ = self._db_adapter._available_.union(self._available_)
        # self._parse_replacements_ is set by metaclass=AdapterMeta
        self._parse_replacements_.update(self._db_adapter._parse_replacements_)

        telemetry.log_api(
            "encapsulate_init",
            config=config,
            additional_props={"is_teleport": self.is_teleport()},
        )

    def submit_python_job(self, *args, **kwargs):
        return super().submit_python_job(*args, **kwargs)

    @available
    @telemetry.log_call(
        "encapsulate_db_materialization", log_args=["materialization"], config=True
    )
    def db_materialization(self, context: dict, materialization: str):
        # NOTE: inspired by https://github.com/dbt-labs/dbt-core/blob/be4a91a0fe35a619587b7a0145e190690e3771c6/core/dbt/task/run.py#L254-L290
        materialization_macro = self.manifest.find_materialization_macro_by_name(
            self.config.project_name, materialization, self._db_adapter.type()
        )

        # HACK: run the entire SQL materialization and return the resulting dict with relations created
        return MacroGenerator(
            materialization_macro, context, stack=context["context_macro_stack"]
        )()

    @property
    @cache_static
    def manifest(self):
        return ManifestLoader.get_full_manifest(self.config)

    def type(self):
        # NOTE: This does not let `fal__` macros to be used
        # Maybe for 1.5 we will get a more reliable way to detect if we are in a SQL or Python context
        if find_funcs_in_stack({"render", "db_materialization"}):
            return self._db_adapter.type()

        return "fal"

    def __getattr__(self, name):
        """
        Directly proxy to the DB adapter, Python adapter in this case does what we explicitly define in this class.
        """
        if hasattr(self._db_adapter, name):
            return getattr(self._db_adapter, name)
        else:
            getattr(super(), name)


def find_funcs_in_stack(funcs: Set[str]) -> bool:
    import inspect

    # NOTE: from https://stackoverflow.com/a/42636264/1276441
    frame = inspect.currentframe()
    while frame:
        if frame.f_code.co_name in funcs:
            return True
        frame = frame.f_back

    return False



from typing import Any, Dict
from dbt.adapters.base import BaseAdapter, Credentials
from trino.sqlalchemy import URL
from dbt.adapters.trino.connections import TrinoCredentials
import sqlalchemy

def create_engine(adapter: BaseAdapter) -> Any:
    creds = adapter.config.credentials

    connect_args = _build_connect_args(creds)

    url = URL(
        host=creds.host,
        port=creds.port,
        catalog=creds.database,
        user=creds.user
    )
    return sqlalchemy.create_engine(url, connect_args=connect_args)

def _build_connect_args(credentials: TrinoCredentials) -> Dict[str, Any]:
    # See:
    # https://github.com/starburstdata/dbt-trino/blob/master/dbt/adapters/trino/connections.py
    return {
        "auth": credentials.trino_auth(),
        "http_scheme": credentials.http_scheme.value
    }



from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, NewType, Optional
from functools import partial
import functools
import zipfile
from dbt.adapters.fal_experimental.connections import TeleportTypeEnum
from dbt.adapters.fal_experimental.utils.environments import EnvironmentDefinition, get_default_pip_dependencies
from dbt.adapters.fal_experimental.utils import extra_path, get_fal_scripts_path, retrieve_symbol
from dbt.config.runtime import RuntimeConfig
from koldstart import KoldstartHost, isolated
import pandas as pd

from dbt.contracts.connection import AdapterResponse

from dbt.fal.adapters.teleport.info import TeleportInfo


DataLocation = NewType('DataLocation', Dict[str, str])

def _prepare_for_teleport(function: Callable, teleport: TeleportInfo, locations: DataLocation) -> Callable:
    @functools.wraps(function)
    def wrapped(relation: str, *args, **kwargs) -> Any:
        relation = relation.lower()
        return function(teleport, locations, relation, *args, **kwargs)

    return wrapped

def _teleport_df_from_external_storage(teleport_info: TeleportInfo, locations: DataLocation, relation: str) -> pd.DataFrame:
    if relation not in locations:
        raise RuntimeError(f"Could not find url for '{relation}' in {locations}")

    if teleport_info.format == 'parquet':
        relation_path = locations[relation]
        url = teleport_info.build_url(relation_path)
        storage_options = _build_teleport_storage_options(teleport_info)
        return pd.read_parquet(url, storage_options=storage_options)
    else:
        # TODO: support more
        raise RuntimeError(f"Format {teleport_info.format} not supported")

def _teleport_df_to_external_storage(teleport_info: TeleportInfo, locations: DataLocation, relation: str, data: pd.DataFrame):
    if teleport_info.format == 'parquet':
        relation_path = teleport_info.build_relation_path(relation)
        url = teleport_info.build_url(relation_path)
        storage_options = _build_teleport_storage_options(teleport_info)

        data.to_parquet(url, storage_options=storage_options)
        locations[relation] = relation_path
        return relation_path
    else:
        raise RuntimeError(f"Format {teleport_info.format} not supported")

def _build_teleport_storage_options(teleport_info: TeleportInfo) -> Dict[str, str]:
    storage_options = {}
    if teleport_info.credentials.type == TeleportTypeEnum.REMOTE_S3:
        storage_options = {
            "key": teleport_info.credentials.s3_access_key_id,
            "secret": teleport_info.credentials.s3_access_key
        }
    elif teleport_info.credentials.type == TeleportTypeEnum.LOCAL:
        pass
    else:
        raise RuntimeError(f"Teleport storage type {teleport_info.credentials.type} not supported")
    return storage_options

def run_with_teleport(
        code: str,
        teleport_info: TeleportInfo,
        locations: DataLocation,
        config: RuntimeConfig,
        local_packages: Optional[bytes]) -> str:
    # main symbol is defined during dbt-fal's compilation
    # and acts as an entrypoint for us to run the model.
    import io
    fal_scripts_path = str(get_fal_scripts_path(config))
    if local_packages is not None:
        if fal_scripts_path.exists():
            import shutil
            shutil.rmtree(fal_scripts_path)
        fal_scripts_path.parent.mkdir(parents=True, exist_ok=True)
        zip_file = zipfile.ZipFile(io.BytesIO(local_packages))
        zip_file.extractall(fal_scripts_path)

    with extra_path(fal_scripts_path):
        main = retrieve_symbol(code, "main")
        return main(
            read_df=_prepare_for_teleport(_teleport_df_from_external_storage, teleport_info, locations),
            write_df=_prepare_for_teleport(_teleport_df_to_external_storage, teleport_info, locations)
        )

def run_in_environment_with_teleport(
    environment: EnvironmentDefinition,
    code: str,
    teleport_info: TeleportInfo,
    locations: DataLocation,
    config: RuntimeConfig,
    adapter_type: str,
) -> AdapterResponse:
    """Run the 'main' function inside the given code on the
    specified environment.

    The environment_name must be defined inside fal_project.yml file
    in your project's root directory."""
    compressed_local_packages = None
    is_remote = type(environment.host) is KoldstartHost

    deps = get_default_pip_dependencies(
        is_remote=is_remote,
        adapter_type=adapter_type,
        is_teleport=True,
    )

    fal_scripts_path = get_fal_scripts_path(config)

    if is_remote and fal_scripts_path.exists():
        with NamedTemporaryFile() as temp_file:
            with zipfile.ZipFile(
                temp_file.name, "w", zipfile.ZIP_DEFLATED
            ) as zip_file:
                for entry in fal_scripts_path.rglob("*"):
                    zip_file.write(entry, entry.relative_to(fal_scripts_path))

            compressed_local_packages = temp_file.read()

    execute_model = partial(
        run_with_teleport,
        code,
        teleport_info,
        locations,
        config,
        compressed_local_packages)

    if environment.kind == "virtualenv":
        requirements = environment.config.get("requirements", [])
        requirements += deps
        isolated_function = isolated(
            kind="virtualenv",
            host=environment.host,
            requirements=requirements
        )(execute_model)
    elif environment.kind == "conda":
        dependencies = environment.config.pop("packages", [])
        dependencies.append({"pip": deps})
        env_dict = {
            "name": "dbt_fal_env",
            "channels": ["conda-forge", "defaults"],
            "dependencies": dependencies
        }
        isolated_function = isolated(
            kind="conda",
            host=environment.host,
            env_dict=env_dict)(execute_model)
    else:
        # We should not reach this point, because environment types are validated when the
        # environment objects are created (in utils/environments.py).
        raise Exception(f"Environment type not supported: {environment.kind}")

    # Machine type is only applicable in KoldstartHost
    if is_remote:
        isolated_function = isolated_function.on(machine_type=environment.machine_type)

    result = isolated_function()
    return result



from dbt.adapters.base import BaseAdapter, BaseRelation
from dbt.adapters.base.connections import AdapterResponse
from dbt.adapters.fal_experimental.adapter_support import new_connection
import pandas as pd
from dbt.adapters.sql import SQLAdapter
import duckdb


def read_relation_as_df(adapter: BaseAdapter, relation: BaseRelation) -> pd.DataFrame:
    db_path = adapter.config.credentials.path

    con = duckdb.connect(database=db_path)
    df = con.execute(f"SELECT * FROM {relation.schema}.{relation.identifier}").fetchdf()
    return df


def write_df_to_relation(
    adapter: SQLAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
) -> AdapterResponse:

    db_path = adapter.config.credentials.path    
    con = duckdb.connect(database=db_path)
    rows_affected = con.execute(
        f"CREATE OR REPLACE TABLE {relation.schema}.{relation.identifier} AS SELECT * FROM data;"
    ).fetchall()[0][0]
    return AdapterResponse("OK", rows_affected=rows_affected)



"""
dbt-fal telemetry code uses source code from: https://github.com/ploomber/ploomber
"""

import datetime
import http.client as httplib
import warnings
import posthog
import pkg_resources
import yaml
import os
from pathlib import Path
import sys
import uuid
from functools import wraps
from typing import Any, List, Optional
import inspect
from dbt.config.runtime import RuntimeConfig

import platform

import atexit

TELEMETRY_VERSION = "0.0.1"
DEFAULT_HOME_DIR = "~/.fal"
CONF_DIR = "stats"
FAL_HOME_DIR = os.getenv("FAL_HOME_DIR", DEFAULT_HOME_DIR)

posthog.project_api_key = "phc_Yf1tsGPPb4POvqVjelT3rPPv2c3FH91zYURyyL30Phy"

invocation_id = uuid.uuid4()


def shutdown():
    posthog.shutdown()
    # HACK: while https://github.com/PostHog/posthog-python/pull/52 happens
    from posthog.request import _session as posthog_session

    posthog_session.close()


atexit.register(shutdown)


def str_param(item: Any) -> str:
    if not isinstance(item, str):
        raise TypeError(f"Variable not supported/wrong type: {item} should be a str")
    return item


def opt_str_param(item: Any) -> Optional[str]:
    if item is None:
        return item
    return str_param(item)


def python_version():
    py_version = sys.version_info
    return f"{py_version.major}.{py_version.minor}.{py_version.micro}"


def is_online():
    """Check if host is online"""
    conn = httplib.HTTPSConnection("www.google.com", timeout=1)

    try:
        conn.request("HEAD", "/")
        return True
    except Exception:
        return False
    finally:
        conn.close()


# Will output if the code is within a container
def is_docker():
    try:
        cgroup = Path("/proc/self/cgroup")
        docker_env = Path("/.dockerenv")
        return (
            docker_env.exists()
            or cgroup.exists()
            and any("docker" in line for line in cgroup.read_text().splitlines())
        )
    except OSError:
        return False


def is_github():
    """Return True if inside a GitHub Action"""
    return os.getenv("GITHUB_ACTIONS") is not None


def is_gitlab():
    """Return True if inside a GitLab CI"""
    return os.getenv("GITLAB_CI") is not None


def get_os():
    """
    The function will output the client platform
    """
    return platform.system()


def dbt_installed_version():
    """Returns: dbt version"""
    try:
        return pkg_resources.get_distribution("dbt-core").version
    except ImportError:
        return


def get_dbt_adapter_type(config: RuntimeConfig) -> str:
    """Returns: the configured actual DBT adapter"""
    target = config.to_target_dict()
    return target["type"]


def fal_installed_version():
    """Returns: fal version"""
    try:
        return pkg_resources.get_distribution("fal").version
    except ImportError:
        return


def is_airflow():
    """Returns: True for Airflow env"""
    return "AIRFLOW_CONFIG" in os.environ or "AIRFLOW_HOME" in os.environ


def get_home_dir():
    """
    Checks if fal home was set through the env variable.
    returns the actual home_dir path.
    """
    return FAL_HOME_DIR if FAL_HOME_DIR else DEFAULT_HOME_DIR


def check_dir_exist(input_location=None):
    """
    Checks if a specific directory exists, creates if not.
    In case the user didn't set a custom dir, will turn to the default home
    """
    home_dir = get_home_dir()

    if input_location:
        p = Path(home_dir, input_location)
    else:
        p = Path(home_dir)

    p = p.expanduser()

    if not p.exists():
        p.mkdir(parents=True)

    return p


def check_uid():
    """
    Checks if local user id exists as a uid file, creates if not.
    """
    uid_path = Path(check_dir_exist(CONF_DIR), "uid.yaml")
    conf = read_conf_file(uid_path)  # file already exist due to version check
    if "uid" not in conf.keys():
        uid = str(uuid.uuid4())
        err = write_conf_file(uid_path, {"uid": uid}, error=True)
        if err:
            return "NO_UID", err, True
        else:
            return uid, None, True

    return conf.get("uid") or "NO_UID", None, False


def check_stats_enabled() -> bool:
    """
    Check if the user allows us to use telemetry. In order of precedence:
    1. If FAL_STATS_ENABLED is defined, check its value
    2. If DO_NOT_TRACK is defined, check its value
    3. Otherwise use the value in stats_enabled in the config.yaml file
    """
    if "FAL_STATS_ENABLED" in os.environ:
        val = os.environ["FAL_STATS_ENABLED"].lower().strip()
        return val != "0" and val != "false" and val != ""

    if "DO_NOT_TRACK" in os.environ:
        val = os.environ["DO_NOT_TRACK"].lower().strip()
        return val != "1" and val != "true"

    # Check if local config exists
    config_path = Path(check_dir_exist(CONF_DIR), "config.yaml")
    if not config_path.exists():
        write_conf_file(config_path, {"stats_enabled": True})
        return True
    else:  # read and return config
        conf = read_conf_file(config_path)
        return conf.get("stats_enabled", True)


def check_first_time_usage():
    """
    The function checks for first time usage if the conf file exists and the
    uid file doesn't exist.
    """
    config_path = Path(check_dir_exist(CONF_DIR), "config.yaml")
    uid_path = Path(check_dir_exist(CONF_DIR), "uid.yaml")
    uid_conf = read_conf_file(uid_path)
    return config_path.exists() and "uid" not in uid_conf.keys()


def read_conf_file(conf_path):
    try:
        with conf_path.open("r") as file:
            conf = yaml.safe_load(file)
            return conf
    except Exception as e:
        warnings.warn(f"Can't read config file {e}")
        return {}


def write_conf_file(conf_path, to_write, error=None):
    try:  # Create for future runs
        with conf_path.open("w") as file:
            yaml.dump(to_write, file)
    except Exception as e:
        warnings.warn(f"Can't write to config file: {e}")
        if error:
            return e


def log_api(
    action: str,
    total_runtime=None,
    config=None,
    additional_props: Optional[dict] = None,
):
    """
    This function logs through an API call, assigns parameters if missing like
    timestamp, event id and stats information.
    """

    if not check_stats_enabled():
        return

    if not is_online():
        return

    additional_props = additional_props or {}

    event_id = uuid.uuid4()

    client_time = datetime.datetime.now()

    uid, uid_error, is_install = check_uid()

    if "NO_UID" in uid:
        additional_props["uid_issue"] = str(uid_error) if uid_error is not None else ""

    config_hash = ""
    if config is not None and hasattr(config, "hashed_name"):
        config_hash = str(config.hashed_name())

    opt_str_param(uid)
    str_param(action)

    props = {
        "tool": "dbt-fal",
        "config_hash": config_hash,
        "event_id": str(event_id),
        "invocation_id": str(invocation_id),
        "user_id": uid,
        "action": action,
        "client_time": str(client_time),
        "total_runtime": str(total_runtime),
        "python_version": python_version(),
        "dbt_version": dbt_installed_version(),
        "dbt_adapter": get_dbt_adapter_type(config),
        "docker_container": is_docker(),
        "airflow": is_airflow(),
        "github_action": is_github(),
        "gitlab_ci": is_gitlab(),
        "os": get_os(),
        "telemetry_version": TELEMETRY_VERSION,
        "$geoip_disable": True,  # This disables GeoIp despite the backend setting
        "$ip": None,  # This disables IP tracking
    }

    all_props = {**props, **additional_props}

    if "argv" in all_props:
        all_props["argv"] = _clean_args_list(all_props["argv"])

    if is_install:
        posthog.capture(distinct_id=uid, event="install_success", properties=all_props)

    posthog.capture(distinct_id=uid, event=action, properties=all_props)


def log_call(action, log_args: List[str] = [], config: bool = False):
    """Runs a function and logs it"""

    def _log_call(func):
        @wraps(func)
        def wrapper(*func_args, **func_kwargs):

            sig = inspect.signature(func).bind(*func_args, **func_kwargs)
            sig.apply_defaults()
            log_args_props = dict(
                map(lambda arg: (arg, sig.arguments.get(arg)), log_args)
            )

            func_self = func_args[0]
            # Get the dbt config from the self object of the method
            dbt_config = func_self.config if config else None

            log_api(
                action=f"{action}_started",
                additional_props={
                    "argv": sys.argv,
                    "args": log_args_props,
                },
                config=dbt_config,
            )

            start = datetime.datetime.now()

            try:
                result = func(*func_args, **func_kwargs)
            except Exception as e:
                log_api(
                    action=f"{action}_error",
                    total_runtime=str(datetime.datetime.now() - start),
                    additional_props={
                        "exception": str(type(e)),
                        "argv": sys.argv,
                        "args": log_args_props,
                    },
                    config=dbt_config,
                )
                raise
            else:
                log_api(
                    action=f"{action}_success",
                    total_runtime=str(datetime.datetime.now() - start),
                    additional_props={
                        "argv": sys.argv,
                        "args": log_args_props,
                    },
                    config=dbt_config,
                )

            return result

        return wrapper

    return _log_call


def _clean_args_list(args: List[str]) -> List[str]:
    ALLOWLIST = [
        "--disable-logging",
        "--project-dir",
        "--profiles-dir",
        "--defer",
        "--threads",
        "--thread",
        "--state",
        "--full-refresh",
        "-s",
        "--select",
        "-m",
        "--models",
        "--model",
        "--exclude",
        "--selector",
        "--all",
        "run",
        "dbt",
        "-v",
        "--version",
        "--debug",
        "--vars",
        "--var",
        "--target",
        "build",
        "clean",
        "compile",
        "debug",
        "deps",
        "docs",
        "init",
        "list",
        "parse",
        "seed",
        "snapshot",
        "source",
        "test",
        "rpc",
        "run-operation",
    ]
    REDACTED = "[REDACTED]"
    output = []
    for item in args:
        if item in ALLOWLIST:
            output.append(item)
        else:
            output.append(REDACTED)
    return output



from pathlib import Path
from dataclasses import dataclass
from typing import Any, Union

from dbt.adapters.base.relation import BaseRelation
from dbt.contracts.relation import ComponentName

@dataclass
class TeleportInfo:
    format: str
    credentials: Any

    @classmethod
    def relation_name(cls, relation: Union[str, BaseRelation]):
        if isinstance(relation, str):
            # TODO: should we check for quoting?
            return relation
        else:
            path = relation.path
            db = path.get_lowered_part(ComponentName.Database)
            sc = path.get_lowered_part(ComponentName.Schema)
            tb = path.get_lowered_part(ComponentName.Identifier)
            return f"{db}.{sc}.{tb}"

    def build_relation_path(self, relation: Union[str, BaseRelation]):
        rel_name = TeleportInfo.relation_name(relation)
        return rel_name + "." + self.format

    def build_url(self, path: str) -> str:
        raise NotImplemented

@dataclass
class LocalTeleportInfo(TeleportInfo):
    base_dir: Path

    def __init__(self, format: str, credentials: Any, base_dir: Union[str, Path]):
        super().__init__(format, credentials)

        self.base_dir = Path(base_dir)
        self.base_dir.mkdir(exist_ok=True)
        assert self.base_dir.is_dir(), "Local Teleport base_dir should be a directory"

    def build_url(self, path: str):
        return self.base_dir / path

# TODO: How to do Teleport Configuration
# Will each adapter have to implement a specific case for each teleport backend.
# And how will we generalize these configurations?
#
# Each adapter will have a specific `teleport` property.
#       my_db:
#         type: duckdb
#         path: ./local_file.db
#         teleport:
#             type: s3
#             bucket: my_bucket
#             access_key_id: ...
#             secret_access_key: ...
#       my_py:
#         type: fal
#         teleport:
#             type: s3
#             s3_bucket: my_bucket
#             s3_access_key_id: ...
#             s3_access_key: ...
# Since all adapters could use different properties and methods to achieve this
@dataclass
class S3TeleportInfo(TeleportInfo):
    bucket: str
    inner_path: str

    def build_url(self, path: str):
        return f's3://{self.bucket}/{self.inner_path}/{path}'



import abc
from contextlib import contextmanager
import time
from typing import (
    Optional,
    Type,
    Dict,
    Any,
    Mapping,
    Iterator,
)

from dbt.adapters.factory import get_adapter
from dbt.exceptions import (
    NotImplementedError,
    DbtRuntimeError,
)
from dbt.contracts.graph.nodes import ResultNode
from dbt.adapters.protocol import AdapterConfig, ConnectionManagerProtocol
from dbt.events.functions import fire_event
from dbt.events.types import (
    CodeExecution,
    CodeExecutionStatus,
)
from dbt.adapters.base.meta import AdapterMeta, available
from dbt.contracts.connection import Credentials, Connection, AdapterResponse


def log_code_execution(code_execution_function):
    # decorator to log code and execution time
    if code_execution_function.__name__ != "submit_python_job":
        raise ValueError("this should be only used to log submit_python_job now")

    def execution_with_log(*args):
        self = args[0]
        connection_name = self.connections.get_thread_connection().name
        fire_event(CodeExecution(conn_name=connection_name, code_content=args[2]))
        start_time = time.time()
        response = code_execution_function(*args)
        fire_event(
            CodeExecutionStatus(
                status=response._message, elapsed=round((time.time() - start_time), 2)
            )
        )
        return response

    return execution_with_log


class PythonJobHelper:
    def __init__(self, parsed_model: Dict, credential: Credentials) -> None:
        raise NotImplementedError("PythonJobHelper is not implemented yet")

    def submit(self, compiled_code: str) -> Any:
        raise NotImplementedError(
            "PythonJobHelper submit function is not implemented yet"
        )


class PythonAdapter(metaclass=AdapterMeta):
    """The PythonAdapter provides an abstract base class for Python adapters.

    Adapters must implement the following methods and macros. Some of the
    methods can be safely overridden as a noop, where it makes sense. Those
    methods are marked with a (passable) in their docstrings. Check docstrings
    for type information, etc.

    To implement a macro, implement "${adapter_type}__${macro_name}" in the
    adapter's internal project.

    To invoke a method in an adapter macro, call it on the 'adapter' Jinja
    object using dot syntax.

    To invoke a method in model code, add the @available decorator atop a method
    declaration. Methods are invoked as macros.

    Methods:
        - is_cancelable

        - python_submission_helpers
        - default_python_submission_method
        - generate_python_submission_response
        or
        - submit_python_job

    Macros:

    """

    ConnectionManager: Type[ConnectionManagerProtocol]

    # A set of clobber config fields accepted by this adapter
    # for use in materializations
    AdapterSpecificConfigs: Type[AdapterConfig] = AdapterConfig

    def __init__(self, config):
        self.config = config
        self.connections = self.ConnectionManager(config)

        # HACK: A Python adapter does not have _available_ all the attributes a DB adapter does.
        # Since we use the DB adapter as the storage for the Python adapter, we must proxy to it
        # all the unhandled calls.
        #
        # Another option is to write a PythonAdapter-specific DBWrapper (PythonWrapper?) that is
        # aware of this case. This may be appealing because a _complete_ adapter (DB+Python) would
        # then be more easily used to replace the Python part of any other adapter.

        self._db_adapter = get_adapter(config)
        self.Relation = self._db_adapter.Relation
        self.Column = self._db_adapter.Column

    def cache_added(self, *args, **kwargs):
        return self._db_adapter.cache_added(*args, **kwargs)

    @available.parse_none
    def get_relation(self, *args, **kwargs):
        return self._db_adapter.get_relation(*args, **kwargs)

    @classmethod
    def date_function(cls):
        # HACK: to appease the ProviderContext
        return """
        import datetime
        return datetime.datetime.now()
        """

    ###
    # Methods that pass through to the connection manager
    ###
    def acquire_connection(self, name=None) -> Connection:
        return self.connections.set_connection_name(name)

    def release_connection(self) -> None:
        self.connections.release()

    def cleanup_connections(self) -> None:
        self.connections.cleanup_all()

    def nice_connection_name(self) -> str:
        conn = self.connections.get_if_exists()
        if conn is None or conn.name is None:
            return "<None>"
        return conn.name

    @contextmanager
    def connection_named(
        self, name: str, node: Optional[ResultNode] = None
    ) -> Iterator[None]:
        try:
            self.acquire_connection(name)
            yield
        finally:
            self.release_connection()

    @contextmanager
    def connection_for(self, node: ResultNode) -> Iterator[None]:
        with self.connection_named(node.unique_id, node):
            yield

    @classmethod
    @abc.abstractmethod
    def is_cancelable(cls) -> bool:
        raise NotImplementedError(
            "`is_cancelable` is not implemented for this adapter!"
        )

    ###
    # Methods that should never be overridden
    ###
    @classmethod
    def type(cls) -> str:
        """Get the type of this adapter. Types must be class-unique and
        consistent.

        :return: The type name
        :rtype: str
        """
        return cls.ConnectionManager.TYPE

    ###
    # ODBC FUNCTIONS -- these should not need to change for every adapter,
    #                   although some adapters may override them
    ###
    def cancel_open_connections(self):
        """Cancel all open connections."""
        return self.connections.cancel_open()

    def pre_model_hook(self, config: Mapping[str, Any]) -> Any:
        """A hook for running some operation before the model materialization
        runs. The hook can assume it has a connection available.

        The only parameter is a configuration dictionary (the same one
        available in the materialization context). It should be considered
        read-only.

        The pre-model hook may return anything as a context, which will be
        passed to the post-model hook.
        """
        pass

    def post_model_hook(self, config: Mapping[str, Any], context: Any) -> None:
        """A hook for running some operation after the model materialization
        runs. The hook can assume it has a connection available.

        The first parameter is a configuration dictionary (the same one
        available in the materialization context). It should be considered
        read-only.

        The second parameter is the value returned by pre_mdoel_hook.
        """
        pass

    def get_compiler(self):
        from dbt.compilation import Compiler

        return Compiler(self.config)

    ###
    # Python
    ###
    @property
    def python_submission_helpers(self) -> Dict[str, Type[PythonJobHelper]]:
        raise NotImplementedError("python_submission_helpers is not specified")

    @property
    def default_python_submission_method(self) -> str:
        raise NotImplementedError("default_python_submission_method is not specified")

    @log_code_execution
    def submit_python_job(
        self, parsed_model: dict, compiled_code: str
    ) -> AdapterResponse:
        submission_method = parsed_model["config"].get(
            "submission_method", self.default_python_submission_method
        )
        if submission_method not in self.python_submission_helpers:
            raise NotImplementedError(
                "Submission method {} is not supported for current adapter".format(
                    submission_method
                )
            )
        job_helper = self.python_submission_helpers[submission_method](
            parsed_model, self.connections.credentials
        )
        submission_result = job_helper.submit(compiled_code)
        # process submission result to generate adapter response
        return self.generate_python_submission_response(submission_result)

    def generate_python_submission_response(
        self, submission_result: Any
    ) -> AdapterResponse:
        raise NotImplementedError(
            "Your adapter need to implement generate_python_submission_response"
        )

    def valid_incremental_strategies(self):
        """The set of standard builtin strategies which this adapter supports out-of-the-box.
        Not used to validate custom strategies defined by end users.
        """
        return ["append"]

    def builtin_incremental_strategies(self):
        return ["append", "delete+insert", "merge", "insert_overwrite"]

    @available.parse_none
    def get_incremental_strategy_macro(self, model_context, strategy: str):
        # Construct macro_name from strategy name
        if strategy is None:
            strategy = "default"

        # validate strategies for this adapter
        valid_strategies = self.valid_incremental_strategies()
        valid_strategies.append("default")
        builtin_strategies = self.builtin_incremental_strategies()
        if strategy in builtin_strategies and strategy not in valid_strategies:
            raise DbtRuntimeError(
                f"The incremental strategy '{strategy}' is not valid for this adapter"
            )

        strategy = strategy.replace("+", "_")
        macro_name = f"get_incremental_{strategy}_sql"
        # The model_context should have MacroGenerator callable objects for all macros
        if macro_name not in model_context:
            raise DbtRuntimeError(
                'dbt could not find an incremental strategy macro with the name "{}" in {}'.format(
                    macro_name, self.config.project_name
                )
            )

        # This returns a callable macro
        return model_context[macro_name]



from dbt.adapters.base import BaseAdapter, BaseRelation
from dbt.adapters.base.connections import AdapterResponse
from dbt.adapters.fal_experimental.adapter_support import new_connection
import pandas as pd

# [bigquery] extras dependencies
import google.cloud.bigquery as bigquery
from google.cloud.bigquery.job import WriteDisposition

from dbt.adapters.bigquery import BigQueryAdapter, BigQueryConnectionManager

def read_relation_as_df(adapter: BaseAdapter, relation: BaseRelation) -> pd.DataFrame:
    sql = f"SELECT * FROM {relation}"

    assert adapter.type() == "bigquery"

    with new_connection(adapter, "fal-bigquery:read_relation_as_df") as conn:

        connection_manager: BaseConnectionManager = adapter.connections  # type: ignore
        client: bigquery.Client = connection_manager.get_thread_connection().handle  # type: ignore

        job = client.query(sql)
        df = job.to_dataframe()

    return df


def write_df_to_relation(
    adapter: BigQueryAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
) -> AdapterResponse:

    assert adapter.type() == "bigquery"

    project: str = relation.database  # type: ignore
    dataset: str = relation.schema  # type: ignore
    table: str = relation.identifier  # type: ignore

    with new_connection(adapter, "fal-bigquery:write_df_to_relation") as conn:
        connection_manager: BigQueryConnectionManager = adapter.connections
        client: bigquery.Client = conn.handle

        table_ref = bigquery.TableReference(
            bigquery.DatasetReference(project, dataset), table
        )

        job_config = bigquery.LoadJobConfig(
            # Specify a (partial) schema. All columns are always written to the
            # table. The schema is used to assist in data type definitions.
            schema=[
                # TODO: offer as a dbt.config parameter?
                # bigquery.SchemaField.from_api_repr(field)
                # for field in (fields_schema or [])
            ],
            source_format="PARQUET",
            write_disposition=WriteDisposition.WRITE_TRUNCATE,
        )

        with connection_manager.exception_handler("START JOB"):
            job = client.load_table_from_dataframe(
                data, table_ref, job_config=job_config
            )

        timeout = connection_manager.get_job_execution_timeout_seconds(conn) or 300

        with connection_manager.exception_handler("LOAD TABLE"):
            adapter.poll_until_job_completes(job, timeout)

        query_table = client.get_table(job.destination)
        num_rows = query_table.num_rows

    # TODO: better AdapterResponse
    return AdapterResponse("OK", rows_affected=num_rows)



from dbt.adapters.base import AdapterPlugin

from dbt.adapters.fal.connections import FalEncCredentials
from dbt.adapters.fal.impl import FalEncAdapter
from dbt.include import fal

Plugin = AdapterPlugin(
    adapter=FalEncAdapter,
    credentials=FalEncCredentials,
    include_path=fal.PACKAGE_PATH
)



from __future__ import annotations

from contextlib import contextmanager
from typing import Iterator

from dbt.adapters.base.impl import BaseAdapter
from dbt.adapters.base.meta import AdapterMeta, available
from dbt.adapters.base.relation import BaseRelation
from dbt.contracts.connection import AdapterResponse

from dbt.fal.adapters.teleport.info import (
    TeleportInfo,
    S3TeleportInfo,
    LocalTeleportInfo,
)
from dbt.fal.adapters.teleport.impl import TeleportAdapter
from dbt.fal.adapters.python.impl import PythonAdapter
from dbt.parser.manifest import MacroManifest, Manifest, ManifestLoader

from . import telemetry

from .connections import FalConnectionManager, FalCredentials, TeleportTypeEnum

from .teleport_adapter_support import wrap_db_adapter
from .teleport import DataLocation, run_in_environment_with_teleport, run_with_teleport

from .adapter_support import reload_adapter_cache
from .adapter import run_in_environment_with_adapter, run_with_adapter

from .utils.environments import fetch_environment, db_adapter_config


class FalAdapterMixin(TeleportAdapter, metaclass=AdapterMeta):
    ConnectionManager = FalConnectionManager

    def __init__(self, config, db_adapter: BaseAdapter):
        self.config = config
        self._db_adapter = db_adapter

        self._relation_data_location_cache: DataLocation = DataLocation({})
        if self.is_teleport():
            self._wrapper = wrap_db_adapter(self._db_adapter, self.credentials.teleport)

    @classmethod
    def type(cls):
        return "fal_experimental"

    @classmethod
    def storage_formats(cls):
        return ["csv", "parquet"]

    @available
    def is_teleport(self) -> bool:
        return getattr(self.credentials, "teleport", None) is not None

    @property
    def manifest(self) -> Manifest:
        return ManifestLoader.get_full_manifest(self.config)

    @property
    def macro_manifest(self) -> MacroManifest:
        return self._db_adapter.load_macro_manifest()


    @telemetry.log_call("experimental_submit_python_job", config=True)
    def submit_python_job(
        self, parsed_model: dict, compiled_code: str
    ) -> AdapterResponse:
        """Execute the given `compiled_code` in the target environment."""
        environment_name = parsed_model["config"].get(
            "fal_environment",
            self.credentials.default_environment,
        )

        machine_type = parsed_model["config"].get(
            "fal_machine",
            "S",
        )

        environment, is_local = fetch_environment(
            self.config.project_root,
            environment_name,
            machine_type,
            self.credentials
        )

        telemetry.log_api(
            "experimental_submit_python_job_config",
            config=self.config,
            additional_props={
                "is_teleport": self.is_teleport(),
                "environment_is_local": is_local,
            },
        )

        if self.is_teleport():
            # We need to build teleport_info because we read from the external storage,
            # we did not _localize_ the data in `teleport_from_external_storage`
            teleport_info = self._build_teleport_info()
            if is_local:
                result_table_path = run_with_teleport(
                    compiled_code,
                    teleport_info=teleport_info,
                    locations=self._relation_data_location_cache,
                    config=db_adapter_config(self.config)
                )
            else:
                result_table_path = run_in_environment_with_teleport(
                    environment,
                    compiled_code,
                    teleport_info=teleport_info,
                    locations=self._relation_data_location_cache,
                    config=db_adapter_config(self.config),
                    adapter_type=self._db_adapter.type()
                )

            relation = self._db_adapter.Relation.create(
                parsed_model["database"], parsed_model["schema"], parsed_model["alias"]
            )
            self._sync_result_table(relation)

            return AdapterResponse("OK")

        else:
            if is_local:
                return run_with_adapter(compiled_code, self._db_adapter, self.config)

            with self._invalidate_db_cache():
                return run_in_environment_with_adapter(
                    environment,
                    compiled_code,
                    db_adapter_config(self.config),
                    self.manifest,
                    self.macro_manifest,
                    self._db_adapter.type()
                )

    @contextmanager
    def _invalidate_db_cache(self) -> Iterator[None]:
        try:
            yield
        finally:
            # Since executed Python code might alter the database
            # layout, we need to regenerate the relations cache
            # after every time we execute a Python model.
            #
            # TODO: maybe propagate a list of tuples with the changes
            # from the Python runner, so that we can tell the cache
            # manager about what is going on instead of hard-resetting
            # the cache-db.
            reload_adapter_cache(self._db_adapter, self.manifest)

    @property
    def credentials(self):
        python_creds: FalCredentials = self.config.python_adapter_credentials
        # dbt-fal is not configured as a Python adapter,
        # maybe we should raise an error?
        assert python_creds is not None
        return python_creds

    def teleport_from_external_storage(
        self, relation: BaseRelation, relation_path: str, teleport_info: TeleportInfo
    ):
        """
        Store the teleport urls for later use
        """

        rel_name = teleport_info.relation_name(relation)
        self._relation_data_location_cache[rel_name] = relation_path

    def teleport_to_external_storage(
        self, relation: BaseRelation, teleport_info: TeleportInfo
    ):
        # Already in external_storage, we do not have local storage
        # Just return the path
        return teleport_info.build_relation_path(relation)

    # TODO: cache this?
    def _build_teleport_info(self):
        teleport_creds = self.credentials.teleport
        assert teleport_creds

        teleport_format = TeleportAdapter.find_format(self, self._wrapper)

        if teleport_creds.type == TeleportTypeEnum.LOCAL:
            assert teleport_creds.local_path
            return LocalTeleportInfo(
                teleport_format, teleport_creds, teleport_creds.local_path
            )
        elif teleport_creds.type == TeleportTypeEnum.REMOTE_S3:
            assert teleport_creds.s3_bucket
            return S3TeleportInfo(
                teleport_format, teleport_creds, teleport_creds.s3_bucket, "teleport"
            )
        else:
            raise NotImplementedError(
                f"Teleport credentials of type {teleport_creds.type} not supported"
            )

    ######
    # HACK: Following implementations only necessary until dbt-core adds Teleport.
    #####
    @available
    def sync_teleport_relation(self, relation: BaseRelation):
        """
        Internal implementation of sync to avoid dbt-core changes
        """
        teleport_info = self._build_teleport_info()
        data_path = self._wrapper.teleport_to_external_storage(relation, teleport_info)
        self.teleport_from_external_storage(relation, data_path, teleport_info)

    def _sync_result_table(self, relation: BaseRelation):
        """
        Internal implementation of sync to put data back into datawarehouse.
        This is necessary because Teleport is not part of dbt-core.
        Once it is and adapters implement it, we will sync the result table back.
        Instead the other adapter will call `sync_teleport` and it will automatically call
        FalAdapter's `teleport_to_external_storage` and the adapter's `teleport_from_external_storage`.
        """
        teleport_info = self._build_teleport_info()
        data_path = self.teleport_to_external_storage(relation, teleport_info)
        self._wrapper.teleport_from_external_storage(relation, data_path, teleport_info)


class FalAdapter(FalAdapterMixin, PythonAdapter):
    def __init__(self, config):
        PythonAdapter.__init__(self, config)
        FalAdapterMixin.__init__(self, config, self._db_adapter)

        telemetry.log_api(
            "experimental_init",
            config=config,
            additional_props={"is_teleport": self.is_teleport()},
        )

    @classmethod
    def is_cancelable(cls) -> bool:
        # TODO: maybe it is?
        return False



from dataclasses import dataclass
from typing import Optional
import os

from dbt.adapters.base import Credentials
from dbt.dataclass_schema import StrEnum, ExtensibleDbtClassMixin

from dbt.fal.adapters.python import PythonConnectionManager


DEFAULT_HOSTS = {
    "cloud": "api.alpha.fal.ai",
    "cloud-eu": "api.eu.fal.ai",
}


class TeleportTypeEnum(StrEnum):
    LOCAL = "local"
    REMOTE_S3 = "s3"


@dataclass
class TeleportCredentials(ExtensibleDbtClassMixin):
    type: TeleportTypeEnum

    # local
    local_path: Optional[str] = os.getcwd()

    # s3
    s3_bucket: Optional[str] = None
    s3_region: Optional[str] = None
    s3_access_key_id: Optional[str] = None
    s3_access_key: Optional[str] = None


class FalConnectionManager(PythonConnectionManager):
    TYPE = "fal_experimental"

    @classmethod
    def open(cls, connection):
        raise NotImplementedError

    def execute(self, compiled_code: str):
        raise NotImplementedError

    def cancel(self, connection):
        raise NotImplementedError


@dataclass
class FalCredentials(Credentials):
    default_environment: str = "local"
    teleport: Optional[TeleportCredentials] = None
    host: str = ""
    key_secret: str = ""
    key_id: str = ""

    # NOTE: So we are allowed to not set them in profiles.yml
    # they are ignored for now
    database: str = ""
    schema: str = ""

    def __post_init__(self):
        if self.host in list(DEFAULT_HOSTS.keys()):
            self.host = DEFAULT_HOSTS[self.host]

    @property
    def type(self):
        return "fal_experimental"

    def _connection_keys(self):
        return ()



import functools
from time import sleep
from typing import Any

import pandas as pd
import sqlalchemy
from contextlib import contextmanager
from dbt.adapters.base import BaseAdapter, BaseRelation, RelationType
from dbt.adapters.base.connections import AdapterResponse, Connection
from dbt.config import RuntimeConfig
from dbt.parser.manifest import MacroManifest, Manifest, ManifestLoader

from dbt.adapters import factory

_SQLALCHEMY_DIALECTS = {
    "redshift": "redshift+psycopg2",
}


def _get_alchemy_engine(adapter: BaseAdapter, connection: Connection) -> Any:
    # The following code heavily depends on the implementation
    # details of the known adapters, hence it can't work for
    # arbitrary ones.
    adapter_type = adapter.type()

    sqlalchemy_kwargs = {}
    format_url = lambda url: url
    if adapter_type == "trino":
        import dbt.adapters.fal_experimental.support.trino as support_trino

        return support_trino.create_engine(adapter)

    if adapter_type == "redshift":
        # If the given adapter supports the DBAPI (PEP 249), we can
        # use its connection directly for the engine.
        sqlalchemy_kwargs["creator"] = lambda *args, **kwargs: connection.handle
        url = _SQLALCHEMY_DIALECTS.get(adapter_type, adapter_type) + "://"
        url = format_url(url)
    else:
        message = (
            f"dbt-fal does not support {adapter_type} adapter. ",
            f"If you need {adapter_type} support, you can create an issue ",
            "in our GitHub repository: https://github.com/fal-ai/fal. ",
            "We will look into it ASAP.",
        )
        raise NotImplementedError(message)

    return sqlalchemy.create_engine(url, **sqlalchemy_kwargs)


def drop_relation_if_it_exists(adapter: BaseAdapter, relation: BaseRelation) -> None:
    if adapter.get_relation(
        database=relation.database,
        schema=relation.schema,
        identifier=relation.identifier,
    ):
        adapter.drop_relation(relation)


def write_df_to_relation(
    adapter: BaseAdapter,
    relation: BaseRelation,
    dataframe: pd.DataFrame,
    *,
    if_exists: str = "replace",
) -> AdapterResponse:
    """Generic version of the write_df_to_relation. Materialize the given
    dataframe to the targeted relation on the adapter."""

    adapter_type = adapter.type()

    if adapter_type == "snowflake":
        import dbt.adapters.fal_experimental.support.snowflake as support_snowflake

        return support_snowflake.write_df_to_relation(adapter, dataframe, relation)

    elif adapter_type == "bigquery":
        import dbt.adapters.fal_experimental.support.bigquery as support_bq

        return support_bq.write_df_to_relation(adapter, dataframe, relation)

    elif adapter_type == "duckdb":
        import dbt.adapters.fal_experimental.support.duckdb as support_duckdb

        return support_duckdb.write_df_to_relation(adapter, dataframe, relation)

    elif adapter_type == "postgres":
        import dbt.adapters.fal_experimental.support.postgres as support_postgres

        return support_postgres.write_df_to_relation(adapter, dataframe, relation)

    else:
        with new_connection(adapter, "fal:write_df_to_relation") as connection:
            # TODO: this should probably live in the materialization macro.
            temp_relation = relation.replace_path(
                identifier=f"__dbt_fal_temp_{relation.identifier}"
            )
            drop_relation_if_it_exists(adapter, temp_relation)

            alchemy_engine = _get_alchemy_engine(adapter, connection)

            # TODO: probably worth handling errors here an returning
            # a proper adapter response.
            rows_affected = dataframe.to_sql(
                con=alchemy_engine,
                name=temp_relation.identifier,
                schema=temp_relation.schema,
                if_exists=if_exists,
                index=False,
            )
            adapter.cache.add(temp_relation)
            drop_relation_if_it_exists(adapter, relation)
            adapter.rename_relation(temp_relation, relation)
            adapter.commit_if_has_connection()

            return AdapterResponse("OK", rows_affected=rows_affected)


def read_relation_as_df(adapter: BaseAdapter, relation: BaseRelation) -> pd.DataFrame:
    """Generic version of the read_df_from_relation."""

    adapter_type = adapter.type()

    if adapter_type == "snowflake":
        import dbt.adapters.fal_experimental.support.snowflake as support_snowflake

        return support_snowflake.read_relation_as_df(adapter, relation)

    elif adapter_type == "bigquery":
        import dbt.adapters.fal_experimental.support.bigquery as support_bq

        return support_bq.read_relation_as_df(adapter, relation)

    elif adapter_type == "duckdb":
        import dbt.adapters.fal_experimental.support.duckdb as support_duckdb

        return support_duckdb.read_relation_as_df(adapter, relation)

    elif adapter_type == "postgres":
        import dbt.adapters.fal_experimental.support.postgres as support_postgres

        return support_postgres.read_relation_as_df(adapter, relation)

    else:
        with new_connection(adapter, "fal:read_relation_as_df") as connection:
            alchemy_engine = _get_alchemy_engine(adapter, connection)
            return pd.read_sql_table(
                con=alchemy_engine,
                table_name=relation.identifier,
                schema=relation.schema,
            )


def prepare_for_adapter(adapter: BaseAdapter, function: Any) -> Any:
    """Prepare the given function to be used with string-like inputs
    (for relations) on the given adapter."""

    @functools.wraps(function)
    def wrapped(quoted_relation: str, *args, **kwargs) -> Any:
        # HACK: we need to drop the quotes from the relation parts
        # This was introduced in https://github.com/dbt-labs/dbt-core/pull/7115
        # and the recommended solution would be to create a macro `fal__resolve_model_name`
        # but it is not possible thanks a macro resolution error we get by returning the db_adapter type.
        # The overall solution could be to avoid creating a Relation and just passing the string as is to the read/write functions.
        parts = map(
            lambda part: part.strip(adapter.Relation.quote_character),
            [*quoted_relation.split(".")],
        )

        relation = adapter.Relation.create(*parts, type=RelationType.Table)
        return function(adapter, relation, *args, **kwargs)

    return wrapped


def reconstruct_adapter(
    config: RuntimeConfig, manifest: Manifest, macro_manifest: MacroManifest
) -> BaseAdapter:
    from dbt.tracking import do_not_track

    # Prepare the DBT to not to track us.
    do_not_track()

    # Prepare the plugin loading system to handle the adapter.
    factory.load_plugin(config.credentials.type)
    factory.load_plugin(config.python_adapter_credentials.type)
    factory.register_adapter(config)

    # Initialize the adapter
    db_adapter = factory.get_adapter(config)
    db_adapter._macro_manifest_lazy = macro_manifest
    reload_adapter_cache(db_adapter, manifest)

    return db_adapter


def reload_adapter_cache(adapter: BaseAdapter, manifest: Manifest) -> None:
    with new_connection(adapter, "fal:reload_adapter_cache"):
        adapter.set_relations_cache(manifest, True)


@contextmanager
def new_connection(adapter: BaseAdapter, connection_name: str) -> Connection:
    with adapter.connection_named(connection_name):
        yield adapter.connections.get_thread_connection()



from typing import Dict, Any, Tuple

from collections import defaultdict
from contextlib import contextmanager
from dbt.config.profile import Profile
from dbt.adapters.base.impl import BaseAdapter, BaseRelation
from dbt.adapters.factory import FACTORY

# TODO: offer in `from isolate import is_agent`
from isolate.connections.common import is_agent

from .connections import FalEncCredentials
from .wrappers import FalEncAdapterWrapper, FalCredentialsWrapper


@contextmanager
def _release_plugin_lock():
    FACTORY.lock.release()
    try:
        yield
    finally:
        FACTORY.lock.acquire()


def load_profiles_info() -> Tuple[Profile, Dict[str, bool], Dict[str, bool]]:
    import os

    from dbt.config.profile import read_profile
    from dbt.config.project import Project
    from dbt.config.renderer import ProfileRenderer
    from dbt.config.utils import parse_cli_vars
    from dbt.main import _build_base_subparser, _add_common_arguments
    from dbt import flags

    # includes vars, profile, target
    parser = _build_base_subparser()
    _add_common_arguments(parser)
    args, _unknown = parser.parse_known_args()

    # dbt-core does os.chdir(project_dir) before reaching this location
    # from https://github.com/dbt-labs/dbt-core/blob/73116fb816498c4c45a01a2498199465202ec01b/core/dbt/task/base.py#L186
    project_root = os.getcwd()

    # from https://github.com/dbt-labs/dbt-core/blob/19c48e285ec381b7f7fa2dbaaa8d8361374136ba/core/dbt/config/runtime.py#L193-L203
    version_check = bool(flags.VERSION_CHECK)
    partial = Project.partial_load(project_root, verify_version=version_check)

    cli_vars: Dict[str, Any] = parse_cli_vars(getattr(args, "vars", "{}"))
    profile_renderer = ProfileRenderer(cli_vars)
    project_profile_name = partial.render_profile_name(profile_renderer)

    # from https://github.com/dbt-labs/dbt-core/blob/19c48e285ec381b7f7fa2dbaaa8d8361374136ba/core/dbt/config/profile.py#L423-L425
    profile_name = Profile.pick_profile_name(
        getattr(args, "profile", None), project_profile_name
    )
    raw_profiles = read_profile(flags.PROFILES_DIR)
    raw_profile = raw_profiles[profile_name]

    # from https://github.com/dbt-labs/dbt-core/blob/19c48e285ec381b7f7fa2dbaaa8d8361374136ba/core/dbt/config/profile.py#L287-L293
    target_override = getattr(args, "target", None)
    if target_override is not None:
        target_name = target_override
    elif "target" in raw_profile:
        target_name = profile_renderer.render_value(raw_profile["target"])
    else:
        target_name = "default"

    fal_dict = Profile._get_profile_data(raw_profile, profile_name, target_name)
    db_profile = fal_dict.get("db_profile")
    assert db_profile, "fal credentials must have a `db_profile` property set"

    try:
        db_profile = Profile.from_raw_profiles(
            raw_profiles,
            profile_name,
            renderer=profile_renderer,
            target_override=db_profile,
        )

        return (
            db_profile,
            {"threads": "threads" in fal_dict},
            {"threads": args.threads is not None},
        )
    except AttributeError as error:
        if "circular import" in str(error):
            raise AttributeError(
                "Do not wrap a type 'fal' profile with another type 'fal' profile"
            ) from error
        else:
            raise


DB_PROFILE = None
DB_RELATION = BaseRelation
INHERIT_PROPERTIES = defaultdict(lambda: True)

# NOTE: Should this file run on isolate agents? Could we skip it entirely and build a FalEncAdapterWrapper directly?
if not is_agent():
    DB_PROFILE, PROFILE_PROPERTIES, ARGS_PROPERTIES = load_profiles_info()
    DB_RELATION = FACTORY.get_relation_class_by_name(DB_PROFILE.credentials.type)

    for key, val in PROFILE_PROPERTIES.items():
        if val:
            INHERIT_PROPERTIES[key] = False
    for key, val in ARGS_PROPERTIES.items():
        if val:
            INHERIT_PROPERTIES[key] = False


class FalEncAdapter(BaseAdapter):
    Relation = DB_RELATION  # type: ignore

    def __new__(cls, config):
        # There are two different credentials types which can be passed to FalEncAdapter
        # 1. FalEncCredentials
        # 2. FalCredentialsWrapper
        #
        # For the first one, we have to go through parsing the profiles.yml (so that we
        # can obtain the real 'db' credentials). But for the other one, we can just use
        # the bound credentials directly (e.g. in isolated mode where we don't actually
        # have access to the profiles.yml file).

        fal_credentials = config.credentials
        if isinstance(fal_credentials, FalEncCredentials):
            db_credentials = DB_PROFILE.credentials
        else:
            # Since profile construction (in the case above) already registers the
            # adapter plugin for the db type, we need to also mimic that here.
            assert isinstance(fal_credentials, FalCredentialsWrapper)
            db_credentials = fal_credentials._db_creds
            with _release_plugin_lock():
                FACTORY.load_plugin(db_credentials.type)

        # TODO: maybe we can do this better?
        with _release_plugin_lock():
            db_adapter_class = FACTORY.get_adapter_class_by_name(db_credentials.type)
            original_plugin = FACTORY.get_plugin_by_name(fal_credentials.type)
            original_plugin.dependencies = [db_credentials.type]

        config.python_adapter_credentials = fal_credentials
        config.sql_adapter_credentials = db_credentials
        if INHERIT_PROPERTIES["threads"]:
            config.threads = DB_PROFILE.threads

        with _release_plugin_lock():
            # Temporary credentials for register
            config.credentials = config.sql_adapter_credentials
            FACTORY.register_adapter(config)
            config.credentials = FalCredentialsWrapper(config.sql_adapter_credentials)

        return FalEncAdapterWrapper(db_adapter_class, config)

    @classmethod
    def type(cls):
        return "fal"



from dataclasses import dataclass
from dbt.adapters.fal_experimental.connections import FalCredentials


@dataclass
class FalEncCredentials(FalCredentials):
    db_profile: str = ""

    def _connection_keys(self):
        return () + super()._connection_keys()

    @property
    def type(self):
        return "fal"

    @property
    def unique_field(self):
        return self.db_profile



from dbt.adapters.base.impl import BaseAdapter
from dbt.adapters.fal_experimental.connections import TeleportCredentials, TeleportTypeEnum

from dbt.fal.adapters.teleport.impl import TeleportAdapter

def wrap_db_adapter(db_adapter: BaseAdapter, teleport_credentials: TeleportCredentials) -> TeleportAdapter:

    if TeleportAdapter.is_teleport_adapter(db_adapter):
        return db_adapter

    # Wrap the adapter with a custom implementation
    if db_adapter.type() == 'duckdb':
        import dbt.adapters.fal_experimental.teleport_support.duckdb as support_duckdb
        return support_duckdb.DuckDBAdapterTeleport(db_adapter, teleport_credentials)

    if db_adapter.type() == 'snowflake':
        import dbt.adapters.fal_experimental.teleport_support.snowflake as support_snowflake
        if teleport_credentials.type == TeleportTypeEnum.REMOTE_S3:
            return support_snowflake.SnowflakeAdapterTeleport(db_adapter, teleport_credentials)
        else:
            raise RuntimeError("Snowflake teleporting works only with S3.")

    raise NotImplementedError(f"Teleport support has not been implemented for adapter {db_adapter.type()}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dbt',
 'dbt.adapters.fal',
 'dbt.adapters.fal_experimental',
 'dbt.adapters.fal_experimental.support',
 'dbt.adapters.fal_experimental.telemetry',
 'dbt.adapters.fal_experimental.teleport_support',
 'dbt.adapters.fal_experimental.utils',
 'dbt.fal.adapters.python',
 'dbt.fal.adapters.teleport',
 'dbt.include.fal',
 'dbt.include.fal_experimental']

package_data = \
{'': ['*'],
 'dbt.include.fal': ['macros/*', 'macros/materializations/*'],
 'dbt.include.fal_experimental': ['macros/materializations/*']}

install_requires = \
['backports.functools_lru_cache>=1.6.4,<2.0.0',
 'dbt-core>=1.4.0,<1.5',
 'dill>=0.3.5.1',
 'koldstart>=0.6.19,<0.7.0',
 'packaging<22',
 'pandas>=1.3.4,<2.0.0',
 'posthog>=1.4.5,<2.0.0',
 'sqlalchemy>=1.4.41,<2.0.0']

extras_require = \
{'bigquery': ['google-cloud-bigquery[pandas]>=3.5.0,<3.6.0'],
 'redshift': ['sqlalchemy-redshift>=0.8.9,<0.9.0'],
 'snowflake': ['snowflake-connector-python[pandas]>=2.8.0,<2.8.2'],
 'teleport': ['s3fs>=2022.8.2'],
 'trino': ['trino[sqlalchemy]>=0.321.0,<0.322.0']}

setup_kwargs = {
    'name': 'dbt-fal',
    'version': '1.4.3',
    'description': 'Simplest way to run dbt python models.',
    'long_description': '<!-- <base href="https://github.com/fal-ai/fal/blob/-/projects/adapter/" target="_blank" /> -->\n\n# Welcome to dbt-fal 👋\n\ndbt-fal adapter is the ✨easiest✨ way to run your [dbt Python models](https://docs.getdbt.com/docs/building-a-dbt-project/building-models/python-models).\n\nStarting with dbt v1.3, you can now build your dbt models in Python. This leads to some cool use cases that was once difficult to build with SQL alone. Some examples are:\n\n- Using Python stats libraries to calculate stats\n- Building forecasts\n- Building other predictive models such as classification and clustering\n\nThis is fantastic! BUT, there is still one issue though! The developer experience with Snowflake and Bigquery is not great, and there is no Python support for Redshift and Postgres.\n\ndbt-fal provides the best environment to run your Python models that works with all other data warehouses! With dbt-fal, you can:\n\n- Build and test your models locally\n- Isolate each model to run in its own environment with its own dependencies\n- [Coming Soon] Run your Python models in the ☁️ cloud ☁️ with elasticly scaling Python environments.\n- [Coming Soon] Even add GPUs to your models for some heavier workloads such as training ML models.\n\n## Getting Started\n\n### 1. Install dbt-fal\n`pip install dbt-fal[bigquery, snowflake]` *Add your current warehouse here*\n\n### 2. Update your `profiles.yml` and add the fal adapter\n\n```yaml\njaffle_shop:\n  target: dev_with_fal\n  outputs:\n    dev_with_fal:\n      type: fal\n      db_profile: dev_bigquery # This points to your main adapter\n    dev_bigquery:\n      type: bigquery\n      ...\n```\n\nDon\'t forget to point to your main adapter with the `db_profile` attribute. This is how the fal adapter knows how to connect to your data warehouse.\n\n### 3. `dbt run`!\nThat is it! It is really that simple 😊\n\n### 4. [🚨 Cool Feature Alert 🚨] Environment management with dbt-fal\nIf you want some help with environment management (vs sticking to the default env that the dbt process runs in), you can create a fal_project.yml in the same folder as your dbt project and have “named environments”:\n\nIn your dbt project folder:\n```bash\n$ touch fal_project.yml\n\n# Paste the config below\nenvironments:\n  - name: ml\n    type: venv\n    requirements:\n      - prophet\n```\n\nand then in your dbt model:\n\n```bash\n$ vi models/orders_forecast.py\n\ndef model(dbt, fal):\n    dbt.config(fal_environment="ml") # Add this line\n\n    df: pd.DataFrame = dbt.ref("orders_daily")\n```\n\nThe `dbt.config(fal_environment=“ml”)` will give you an isolated clean env to run things in, so you dont pollute your package space.\n\n### 5. [Coming Soon™️] Move your compute to the Cloud!\nLet us know if you are interested in this. We are looking for beta users.\n\n# Have feedback or need help?\n\n- Join us in [fal on Discord](https://discord.com/invite/Fyc9PwrccF)\n- Join the [dbt Community](http://community.getdbt.com/) and go into our [#tools-fal channel](https://getdbt.slack.com/archives/C02V8QW3Q4Q)\n',
    'author': 'Features & Labels',
    'author_email': 'hello@fal.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fal-ai/fal/blob/-/projects/adapter',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<3.12',
}


setup(**setup_kwargs)



import abc
from typing import List, Union

from dbt.exceptions import NotImplementedError
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.base.impl import BaseAdapter

from dbt.fal.adapters.teleport.info import TeleportInfo


class TeleportAdapter:
    """The TeleportAdapter provides an abstract class for adapters usable with Teleport.

    Adapters must implement the following methods and macros. Some of the
    methods can be safely overridden as a noop, where it makes sense. Those
    methods are marked with a (passable) in their docstrings. Check docstrings
    for type information, etc.

    To implement a macro, implement "${adapter_type}__${macro_name}" in the
    adapter's internal project.

    To invoke a method in an adapter macro, call it on the 'adapter' Jinja
    object using dot syntax.

    To invoke a method in model code, add the @available decorator atop a method
    declaration. Methods are invoked as macros.

    Methods:
        - storage_formats
        # - teleport_backends
        - teleport_from_external_storage
        - teleport_to_external_storage

    Macros:
        -

    """

    @classmethod
    @abc.abstractmethod
    def storage_formats(cls) -> List[str]:
        """
        List of formats this adapter handles. e.g. `['csv', 'parquet']`
        """
        raise NotImplementedError(
            "`storage_formats` is not implemented for this adapter!"
        )

    @abc.abstractmethod
    def teleport_from_external_storage(
        self, relation: BaseRelation, data_path: str, teleport_info: TeleportInfo
    ) -> None:
        """
        Localize data found in `data_path` into the a local table with the name given by `relation`.
        Handle each possible format defined in `storage_formats` method.
        """
        raise NotImplementedError(
            "`teleport_from_external_storage` is not implemented for this adapter!"
        )

    @abc.abstractmethod
    def teleport_to_external_storage(
        self, relation: BaseRelation, teleport_info: TeleportInfo
    ) -> str:
        """
        Take local table `relation` and upload the data with Teleport. Return the `data_path` the data is uploaded to.
        Handle each possible format defined in `storage_formats` method.
        """
        raise NotImplementedError(
            "`teleport_to_external_storage` is not implemented for this adapter!"
        )

    @classmethod
    def is_teleport_adapter(
        cls, adapter: Union[BaseAdapter, "TeleportAdapter"]
    ) -> bool:
        methods = [
            "storage_formats",
            # "teleport_backends",
            "teleport_from_external_storage",
            "teleport_to_external_storage",
        ]
        return isinstance(adapter, TeleportAdapter) or all(
            map(lambda m: hasattr(adapter, m), methods)
        )

    @classmethod
    def find_format(
        cls, target_adapter: "TeleportAdapter", ref_adapter: "TeleportAdapter"
    ):
        """
        Find common format between target and ref adapter, giving priority to target list ordering.
        """
        target_formats = target_adapter.storage_formats()
        ref_formats = ref_adapter.storage_formats()
        for format in target_formats:
            if format in ref_formats:
                return format

        raise RuntimeError(
            f"No common format between {target_adapter.type()} and {ref_adapter.type()} "
            f"—  {target_formats} | {ref_formats}"
        )
